#!/usr/bin/env python

# TODO: need to commit updates to emod_task.py

# TODO: Need to demand update to calibra to remove .config.parameters -> .parameters in callback .
# TODO: Need to make PR to calibra for param/plotting related fix (.copy() fix)

# TODO: update code to relocate Malawi_Fertility_Historical.csv to some inputs directory
# TODO: same for Malawi_male_mortality.csv
# TODO: same for Malawi_female_mortality.csv

'''
Requirements:
- local files:
-- control_params.py for setting calibration-run-level controls
-- manifest.py with pathing information for input dirs and files
-- config.py with method set_config(config) for generating base config(.json) object
-- campaign_demog.py with methods build_campaign(params=None) and build_demographics() for building
the campaign(.json) and demographics (json) files
-- custom_sample_handling.py contains map_sample_to_model_input(simulation, sample) for custom mapping as needed
as well as constrain_sample(sample) for custom sample massaging

'''

from functools import partial

from idmtools.assets import AssetCollection
from idmtools.core.platform_factory import Platform

from idmtools_calibra.calib_manager import CalibManager
from idmtools_calibra.algorithms.optim_tool import OptimTool
from idmtools_calibra.plotters.likelihood_plotter import LikelihoodPlotter
from idmtools_calibra.plotters.optim_tool_plotter import OptimToolPlotter

from emodpy.emod_task import EMODTask

from emodpy_hiv.lib.analysis.hiv_calib_site import HIVCalibSite
from emodpy_hiv.lib.utils.project_data import parse_ingest_data_from_xlsm
from emodpy_hiv.lib.utils.runtime import add_post_channel_config_as_asset, create_calibration_name, \
    initialize_config_object, map_sample_to_model_input, set_post_processing, constrain_sample, compute_num_cores

# project-local files
import os
import sys
sys.path.append(os.getcwd())
import project_files as project


def get_ingest_information(ingest_filename):
    # params is a dict, site_info is a dict, reference is a PopulationObs object, analyzers is a list of dictionaries of
    # analyzer arguments
    params, site_info, reference, analyzers, channels = parse_ingest_data_from_xlsm(filename=ingest_filename)
    analyzers = [analyzers[0]]
    # making this available to any script that imports this file as a module, like run_scenarios.py
    ingest_info = {
        'params': params,
        'site_info': site_info,
        'reference': reference,
        'analyzers': analyzers,
        'channels': channels
    }

    site = HIVCalibSite(analyzers=analyzers, site_data=site_info, reference_data=reference, force_apply=True)
    return ingest_info, site


def initialize_config_object_wrapper(config_obj):
    return initialize_config_object(config_obj=config_obj,
                                    control_params=project.control_params,
                                    config_setter=project.config.set_config)


def set_post_processing_wrapper(task):
    return set_post_processing(task=task, path=project.manifest.POST_PROCESSING_PATH)


def initialize_EMODTask(channels, reference, site_info):
    task = EMODTask.from_default2(common_assets=AssetCollection(),
                                  eradication_path=project.manifest.eradication_path,
                                  demog_builder=project.campaign_demog.build_demographics,
                                  schema_path=project.manifest.SCHEMA_PATH,
                                  param_custom_cb=initialize_config_object_wrapper,
                                  ep4_custom_cb=set_post_processing_wrapper)

    # generate channel config as an asset
    add_post_channel_config_as_asset(task, channels, reference, site_info)
    return task


def set_additional_run_time_options(task, max_memory_mb):
    additional_config_parameters2 = {
        'Serialization_Time_Steps': [0],  # TODO: needed, really???? Ask jonathan why 0 by itself is broken but is default!
        'Memory_Usage_Halting_Threshold_Working_Set_MB': max_memory_mb,
        'Memory_Usage_Warning_Threshold_Working_Set_MB': int(0.9 * max_memory_mb)
    }
    for parameter, value in additional_config_parameters2.items():
        setattr(task.config.parameters, parameter, value)


def constrain_sample_wrapper(sample):
    return constrain_sample(sample=sample,
                            custom_sample_constrainer=project.custom_sample_handling.constrain_sample)


def initialize_optim_tool(params, n_samples_per_iteration, center_repeats, volume_fraction):
    parameter_count = len([p for p in params if p['Dynamic']])
    if parameter_count == 0:
        warning_note = \
            "/!\\ WARNING /!\\ OptimTool requires at least one of params with Dynamic set to True. Exiting..."
        print(warning_note)
        exit()

    # Compute hypersphere radius as a function of the number of dynamic parameters
    r = OptimTool.get_r(parameter_count, volume_fraction)

    optim_tool = OptimTool(
        params,
        constrain_sample_wrapper,  # <-- Will not be saved in iteration state
        mu_r=r,  # <-- Mean fraction of parameter range for numerical derivative.  Do not go too low with integer parameters
        sigma_r=r / 10.,  # <-- stddev of above radius
        samples_per_iteration=n_samples_per_iteration,
        center_repeats=center_repeats,
        rsquared_thresh=0.81
        # <-- Linear regression goodness of fit threshold, [0:1].  Above this, regression is used.  Below, use best point. Best to be fairly high.
    )
    return optim_tool


def campaign_builder_wrapper(params=None):
    return project.campaign_demog.build_campaign(schema_path=project.manifest.SCHEMA_PATH,
                                                 control_params=project.control_params, params=params)


def generate_map_sample_to_model_input_wrapper(params, campaign_builder, sample_mapper):
    return partial(map_sample_to_model_input,
                   parameter_knob_dict=params,
                   campaign_builder=campaign_builder,
                   custom_sample_mapper=sample_mapper)


def initialize_calib_manager(task, site, calibration_name, n_replicates, n_iterations, next_point_object,
                             sample_mapping_function):
    calib_manager = CalibManager(
        name=calibration_name,
        task=task,
        map_sample_to_model_input_fn=sample_mapping_function,
        sites=[site],
        next_point=next_point_object,
        sim_runs_per_param_set=n_replicates,
        max_iterations=n_iterations,
        plotters=[LikelihoodPlotter(), OptimToolPlotter()]
    )
    return calib_manager


DEFAULTS = {
    'n_iterations': 2,
    'n_replicates': 1,
    'suffix': 'testing-only',
    'n_samples': 3,
    'n_center_repeats': 2,
    'volume_fraction': 0.01,
    'platform': 'Calculon',
    'max_memory': 8000
}


def parse_args():
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument('-s', '--samples', dest='n_samples', type=int, default=DEFAULTS['n_samples'],
                        help=f"Number of samples to run per calibration iteration (Default: {DEFAULTS['n_samples']})")
    parser.add_argument('-r', '--replicates', dest='n_replicates', type=int, default=DEFAULTS['n_replicates'],
                        help=f"Number of replicates to use for parameter samples (Default: {DEFAULTS['n_replicates']})")
    parser.add_argument('-c', '--center-repeats', dest='n_center_repeats', type=int, default=DEFAULTS['n_center_repeats'],
                        help=f"Number of center repeats to run per iteration (Default: {DEFAULTS['n_center_repeats']})")
    parser.add_argument('-i', '--iterations', dest='n_iterations', type=int, default=DEFAULTS['n_iterations'],
                        help=f"Number iterations (in total) to run (Default: {DEFAULTS['n_iterations']})")
    parser.add_argument('-v', '--volume-fraction', dest='volume_fraction', type=float, default=DEFAULTS['volume_fraction'],
                        help=f"Fraction of parameter space to explore per iteration (Default: {DEFAULTS['volume_fraction']})")
    parser.add_argument('-S', '--suffix', dest='suffix', type=str, default=DEFAULTS['suffix'],
                        help=f"Suffix applied to calibration name (Default: {DEFAULTS['suffix']})")
    parser.add_argument('-p', '--platform', dest='platform', type=str, default=DEFAULTS['platform'],
                        help=f"Platform to run calibration on (Default: {DEFAULTS['platform']})")
    parser.add_argument('-m', '--max-memory', dest='max_memory', type=int, default=DEFAULTS['max_memory'],
                        help=f"Maximum memory of simulations before erroring (MB) (Default: {DEFAULTS['max_memory']})")

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse_args()

    ingest_info, site = get_ingest_information(ingest_filename=project.manifest.INGEST_FILENAME)

    import pathlib # for a join
    import emod_hiv.bootstrap as dtk
    dtk.setup(pathlib.Path(project.manifest.eradication_path).parent)

    task = initialize_EMODTask(channels=ingest_info['channels'],
                               reference=ingest_info['reference'],
                               site_info=ingest_info['site_info'])
    set_additional_run_time_options(task=task, max_memory_mb=args.max_memory)
    num_cores = compute_num_cores(max_memory_mb=args.max_memory)

    sample_mapping_function = generate_map_sample_to_model_input_wrapper(ingest_info['params'],
                                                                         campaign_builder=campaign_builder_wrapper,
                                                                         sample_mapper=project.custom_sample_handling.map_sample_to_model_input)
    calibration_name = create_calibration_name(site_name=site.name,
                                               pop_scale_factor=project.control_params.base_population_scale_factor,
                                               n_replicates=args.n_replicates,
                                               suffix=args.suffix)
    next_point_object = initialize_optim_tool(params=ingest_info['params'],
                                              n_samples_per_iteration=args.n_samples,
                                              center_repeats=args.n_center_repeats,
                                              volume_fraction=args.volume_fraction)
    calib_manager = initialize_calib_manager(task=task,
                                             site=site,
                                             calibration_name=calibration_name,
                                             n_replicates=args.n_replicates,
                                             n_iterations=args.n_iterations,
                                             next_point_object=next_point_object,
                                             sample_mapping_function=sample_mapping_function)
    calib_manager.platform = Platform(args.platform, num_cores=num_cores)

    calib_manager.run_calibration()
